#include <cstdio>
#include <cstdlib>

#include "stream.h"
#include "uint40.h"
#include "utils.h"

int main(int argc, char **argv) {
  if (argc != 3) {
    fprintf(stderr, "%s <infile> <outfile>\n\n"
        "Decode LZ77 encoding stored in <infile>\n"
        "and write into <outwrite>\n", argv[0]);
    std::exit(EXIT_FAILURE);
  }

  stream_reader<uint40> *reader =
    new stream_reader<uint40>(argv[1]);

  long length = reader->read().ll();
  fprintf(stderr, "Input file: %s\n", argv[1]);
  fprintf(stderr, "Output file: %s\n", argv[2]);
  fprintf(stderr, "length = %ld\n", length);

  unsigned char *text = new unsigned char[length];
  long ptr = 0, dbg = 0;

  while (!(reader->empty())) {
    long pos = reader->read().ll();
    long len = reader->read().ll();

    if (!len) text[ptr++] = (unsigned char)pos;
    else {
      for (long j = 0; j < len; ++j)
        text[ptr++] = text[pos + j];
    }

    // Print progress.
    if (pos - dbg > (10 << 20)) {
      dbg = pos;
      fprintf(stderr, "Decompressed %.2LfMiB (%.2Lf%%)\r",
          (long double)pos / (1 << 20),
          (100.L * pos) / length);
    }
  }
  fprintf(stderr, "Decompressed %.2LfMiB (100.00%%)\n",
      (long double)length / (1 << 20));

  std::reverse(text, text + length);

  fprintf(stderr, "Writing to file... ");
  utils::write_text_to_file(text, length, argv[2]);

  fprintf(stderr, "DONE\n");
  delete[] text;
}

