#ifndef __ENDMAP_H
#define __ENDMAP_H

#include <cstdio>
#include <string>
#include <stdint.h>
#include <algorithm>

#include "utils.h"

#define em_bufsize (1 << 18)

struct end_map {
  end_map(std::string fn, long an)
      :fname(fn), n(an) {
    buf = new long[em_bufsize];
    written = 0; // how many 64-bit numbers are in file
    origin = 0; // index of first element in the buffer
    filled = 0; // number of elements in the buffer
    ptr = 0; // pointer to first unread element in the buffer

    f = utils::open_file(fname, "wb+");
    
    // Pad the file with zeros.
    std::fill(buf, buf + em_bufsize, 0L);

    for (long i = 0; i < n / em_bufsize; ++i)
      std::fwrite(buf, sizeof(long), em_bufsize, f);

    if (n % em_bufsize)
      std::fwrite(buf, sizeof(long), n % em_bufsize, f);
  }
  
  long next() {
    if (ptr >= filled) {
      origin += filled;
      filled = std::min((long)em_bufsize, n - origin);
      long r = fread(buf, sizeof(long), filled, f);

      if (r != filled) {
        std::perror("endmap: fread");
        std::exit(EXIT_FAILURE);
      }

      ptr = 0;
    }
 
    return buf[ptr++];
  }

  bool isnext() const {
    if (!written) return false;
    else return (ptr < filled || origin + filled != n);
  }
  
  void write_block(int64_t *b, long length) {
    if (std::fseek(f, 0, SEEK_END)) {
      std::perror("endmap: fseek2");
      std::exit(EXIT_FAILURE);
    }

    if (std::fseek(f, -8 * (written + length), SEEK_CUR)) {
      std::perror("endmap: fseek3");
      std::exit(EXIT_FAILURE);
    }

    std::fwrite(b, sizeof(long), length, f);

    if (std::fseek(f, -8 * length, SEEK_CUR)) {
      std::perror("endmap: fseek4");
      std::exit(EXIT_FAILURE);
    }

    origin = ftell(f) / sizeof(long);
    filled = ptr = 0;
    written += length;
  }

  void rewind() {
    if (fseek(f, 0, SEEK_END)) {
      perror("endmap: fseek2");
      std::exit(EXIT_FAILURE);
    }

    if (fseek(f, -8 * (written), SEEK_CUR)) {
      perror("endmap: fseek3");
      std::exit(EXIT_FAILURE);
    }

    origin = ftell(f) / sizeof(long);
    filled = ptr = 0;
  }

  ~end_map() {
    if (f)
      std::fclose(f);
    delete[] buf;
  }

  std::string fname;

  long *buf;
  long filled, origin, written, ptr;

  FILE *f;
  long n;
};

#endif // __ENDMAP_H
