#ifndef __MATCH_H
#define __MATCH_H

#include <cstring>
#include <algorithm>

#include "reader.h"

// Update MS-factorization of t[0..n-1) (variables s,p)
// to MS-factorizatoin of t[0..n)
void next(unsigned char *t, int &n, int &s, int &p) {
  ++n; if (n == 1) { s = 0; p = 1; return; }
  int i = n - 1, r = (i - s) % p;
  while (i < n) {
    unsigned char a = t[s + r], b = t[i];
    if (a > b) { p = i - s + 1; r = 0; }
    else if (a < b) { i -= r; s = i; p = 1; r = 0; }
    else { ++r; if (r == p) r = 0; } ++i;
  }
}

// Given the MS-factorizarion of t[0..n) return the length of its
// longst border (say, k) and find MS-factorization of P[0..k)
int longest_border(unsigned char *t, int n, int &s, int&p) {
  if (n > 0 && 3 * p <= n && !memcmp(t, t + p, s)) return n - p;
  int i = n / 3 + 1, j = 0;
  while (i < n) {
    while (i + j < n && t[i + j] == t[j]) next(t, j, s, p);
    if (i + j == n) return j;
    if (j > 0 && 3 * p <= j && !memcmp(t, t + p, s)) { i += p, j -= p; }
    else { i += j / 3 + 1; j = 0; }
  }
  return 0;
}

std::pair<long, int> pattern_matching(long end, long n, unsigned char *pat, int m, file_reader &fr) {
  int len = 0, j = 0, s, p;
  long pos = -1, i = 0;
  while (i < end) {
    while (j < m && i + j < n && fr.get(n - 1 - (i + j))/*x[i + j]*/ == pat[j]/*x[beg + j]*/)
      next(pat, j, s, p);
    if (j > len) { len = j; pos = i; }
    int oldj = j; j = longest_border(pat, j, s, p);
    if (!j) ++i; else i += oldj - j;
  }
  return std::make_pair(pos, len);
}


#endif // __MATCH_H
