#ifndef __READER_H
#define __READER_H

#include <cstdio>
#include <stdint.h>
#include <string>
#include <algorithm>

#include "utils.h"

#define fr_bufsize (1 << 18)

struct file_reader {
  file_reader(std::string fn, long an)
      :fname(fn), n(an) {
    buf = new unsigned char[fr_bufsize];
    origin = filled = fpos = 0;

    f = utils::open_file(fname, "rb");
  }

  inline unsigned char get(long i) {
    if (i >= origin && i < origin + filled)  return buf[i - origin];
    else {
      // Refill buffer centered at position i.
      long neworigin = std::max(0L, i - fr_bufsize / 2);
      int toread = std::min(n - neworigin, (long)fr_bufsize);
      long roffset = 0;
      filled = 0;

      // Save some disk I/O if possible.
      if (i < origin && neworigin + toread > origin) {
        int overlap = neworigin + toread - origin;
        for (int j = 1; j <= overlap; ++j)
          buf[toread - j] = buf[overlap - j];
        filled = overlap;
        toread -= overlap;
      } else if (i > origin + filled && neworigin < origin + filled) {
        int overlap = origin + filled - neworigin;
        for (int j = 0; j < overlap; ++j)
          buf[j] = buf[filled - overlap + j];
        filled = overlap;
        toread -= overlap;
        roffset = overlap;
      } else filled = 0;

      // Do the disk I/O.
      long offset = neworigin + roffset - fpos;

      if (offset && std::fseek(f, offset, SEEK_CUR)) {
        std::perror("Error in filer_reader: fseek1");
        std::exit(EXIT_FAILURE);
      }

      if (std::ftell(f) != neworigin) {
        std::perror("Error in file_reader: something is wrong");
        std::exit(EXIT_FAILURE);
      }

      int r = std::fread(buf + roffset, 1, toread, f);

      if (r != toread) {
        std::perror("Error in file_reader: fread1");
        std::exit(EXIT_FAILURE);
      }

      filled += toread;
      fpos = std::ftell(f);
      origin = neworigin;

      return buf[i - origin];
    }
  }

  void read_block(unsigned char *b, int length, long orig) {
    if (std::fseek(f, orig, SEEK_SET)) {
      std::perror("Error in filer_reader: fseek2");
      std::exit(EXIT_FAILURE);
    } // optimize?

    int r = std::fread(b, 1, length, f);
    if (r != length) {
      std::perror("Error in file_reader: fread2");
      std::exit(EXIT_FAILURE);
    }

    origin = filled = 0;
    fpos = std::ftell(f);
  }

  ~file_reader() {
    if (f)
      std::fclose(f);
    delete[] buf;
  }

  std::string fname;
  unsigned char *buf;
  long filled, origin;

  std::FILE *f;
  long n, fpos;
};

#endif // __READER_H
