#ifndef __RMQ_TREE_H
#define __RMQ_TREE_H

#include <algorithm>
#include <vector>

struct rmq_tree {
  rmq_tree(int *ax, int an, int b)
      :x(ax), n(an), size(1), bits(b), pow(1 << bits), mask(pow - 1) {
    while ((size << bits) < n) size <<= 1;
    tree.resize(size << 1 , n);

    // Initialize the tree.
    for (int i = 0, j = 0; i < n; i += pow, ++j)
      tree[size + j] = *std::min_element(x + i, x + std::min(n, i + pow));
    for (int i = size - 1; i >= 1; --i)
      tree[i] = std::min(tree[2 * i], tree[2 * i + 1]);
  }

  // Return largest j < i such that x[j] < ub.
  inline int psv(int i, int ub) {
    // Scan nearby positions.
    int j = i;
    while (i >= 0 && x[i] >= ub && j - i < 512) --i;
    if (i >= 0 && x[i] < ub) return x[i];
    else if (i <= 0) return -1;

    // Scan up to a block boundary.
    for (j = i - 1; (j + 1) & mask; --j)
      if (x[j] < ub) return x[j];

    // Locate the lowest left-neighbor with key < ub.
    for (i = size + (i >> bits); i != 1; i >>= 1)
      if ((i & 1) && tree[i - 1] < ub) { --i; break; }
    if (i == 1) return -1;

    // Narrow the range to a single block and scan it.
    while (i < size) i = (i << 1) + (tree[2 * i + 1] < ub);
    for (i = (i - size) << bits, j = std::min(n, i + pow) - 1; i <= j; --j)
      if (x[j] < ub) return x[j];
    return -1;
  }

  // Analogous to psv.
  inline int nsv(int i, int ub) {
    int j = i;
    while (i < n && x[i] >= ub && i - j < 512) ++i;
    if (i < n && x[i] < ub) return x[i];
    else if (i >= n) return -1;

    for (j = i + 1; j < n && (j & mask); ++j)
      if (x[j] < ub) return x[j];
 
    for (i = size + (i >> bits); i != 1; i >>= 1)
      if (!(i & 1) && tree[i + 1] < ub) { ++i; break; }
    if (i == 1) return -1;

    while (i < size) i = (i << 1) + (tree[2 * i] >= ub);
    for (i = (i - size) << bits, j = std::min(n, i + pow); i < j; ++i)
      if (x[i] < ub) return x[i];
    return -1;
  }

  std::vector<int> tree;
  int *x, n, size, bits;
  const int pow, mask;
};

#endif // __RMQ_TREE_H
